/*-
 * Copyright (c) 2001-2002 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: kip6fw.c,v 1.3.2.5 2002/09/24 10:52:30 simon Exp $";
#endif /* !lint */

#include "kip6fw.h"

#ifdef USE_FreeBSD_IP6FW

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <varargs.h>

#include "config.h"
#include "db.h"
#include "debug.h"
#include "rules.h"

u_quad_t	kip6fw_bcnt_max;
struct ip6_fw	*kip6fw;
u_int		nkip6fw;

static int	sd = -1;
static int	nbytes;
static u_int	nalloc;

static char	ip6fw_errmsg[] = "IPv6 Firewall is not configured in the kernel or does not work properly";


/*
 * Init IPv6 Firewall support: open raw socket.
 */
int
kip6fw_init(void)
{
	struct ip6_fw	tmp;

	syslog(LOG_INFO, "preinit IPv6 Firewall support");
	if ( (sd = socket(AF_INET6, SOCK_RAW, IPPROTO_RAW)) < 0) {
		syslog(LOG_ERR, "kip6fw_init: socket(AF_INET6, SOCK_RAW, IPPROTO_RAW): %m");
		return -1;
	}
	kip6fw = NULL;
	nkip6fw = nalloc = 0;

	switch (sizeof(tmp.fw_bcnt)) {
	case 4:
		kip6fw_bcnt_max = ULONG_MAX;
		break;
	case 8:
		kip6fw_bcnt_max = UQUAD_MAX;
		break;
	default:
		syslog(LOG_ERR, "kip6fw_init: unsupported size of fw_bcnt (%d bytes) in struct ip6_fw",
		    sizeof(tmp.fw_bcnt));
		return -1;
	}

	if (debug_ip6fw > 0) {
		syslog(LOG_INFO, "kip6fw_init: raw socket opened, FD %d", sd);
		if (debug_ip6fw > 1)
			syslog(LOG_INFO, "kip6fw_init: sizeof(struct ip6_fw) = %u, sizeof(fw_bcnt) = %u",
			    sizeof tmp, sizeof tmp.fw_bcnt);
	}

	return 0;
}

/*
 * Read table of IP6FW rules from kernel. Realloc internal structures
 * if needed.
 */
int
kip6fw_read_table(void)
{
	int		ret;
	static u_int	maxbytes = (USHRT_MAX + 1) * sizeof *kip6fw;
	static int	nbytesalloc;

	if (kip6fw != NULL) {
		nbytes = nbytesalloc;
		if (getsockopt(sd, IPPROTO_IPV6, IPV6_FW_GET, kip6fw, &nbytes) < 0) {
			syslog(LOG_ERR, "kip6fw_read_table: getsockopt(IPV6_FW_GET): %m");
			syslog(LOG_ERR, "%s", ip6fw_errmsg);
			return -1;
		}
		nkip6fw = nbytes / sizeof(struct ip6_fw);
		if (debug_ip6fw > 0)
			syslog(LOG_INFO, "kip6fw_read_table: nkip6fw = %u, nbytes = %u, nbytesalloc = %u",
			    nkip6fw, nbytes, nbytesalloc);
	}
	while (nkip6fw >= nalloc) {
		nalloc = nalloc * 2 + 200;
		nbytes = nbytesalloc = nalloc * sizeof(struct ip6_fw);
		if ( (kip6fw = realloc(kip6fw, nbytes)) == NULL) {
			syslog(LOG_ERR, "kip6fw_read_table: realloc(%d bytes): %m", nbytes);
			return -1;
		}
		ret = getsockopt(sd, IPPROTO_IPV6, IPV6_FW_GET, kip6fw, &nbytes);
		if ((ret < 0 && errno != EINVAL) || nbytesalloc > maxbytes) {
			syslog(LOG_ERR, "kip6fw_read_table: getsockopt(IPV6_FW_GET): %m");
			syslog(LOG_ERR, "%s", ip6fw_errmsg);
			return -1;
		}
		nkip6fw = nbytes / sizeof(struct ip6_fw);
		if (debug_ip6fw > 0) {
			syslog(LOG_INFO, "kip6fw_read_table: resize table");
			syslog(LOG_INFO, "kip6fw_read_table: nkip6fw = %u, nbytes = %u, nbytesalloc = %u, nalloc = %u",
			    nkip6fw, nbytes, nbytesalloc, nbytesalloc / sizeof(struct ip6_fw));
		}
	}
	if (debug_ip6fw > 1) {
		u_int	i;
		char	one[10 + 5 + 1 + 1 + 1];
		/* 10 - length of u_int, 5 - length of u_short,
		   1 - '.', 1 - ' ', 1 - '\0' */
		int	len, total_len;
		u_int	subnumber = 0;
#define TXT_LEN 60
		char	txt[TXT_LEN + 1];
		int	prevnum = -1;
		struct ip6_fw	*kip6fwp = kip6fw;

		syslog(LOG_INFO, "kip6fw_read_table: dump IP6FW table (%u record%s):",
		    nkip6fw, nkip6fw > 1 ? "s" : "");
		for (i = 0, total_len = 0; i < nkip6fw; ++kip6fwp, ++i) {
			if (prevnum == kip6fwp->fw_number)
				++subnumber;
			else {
				prevnum = kip6fwp->fw_number;
				subnumber = 0;
			}
			if (subnumber == 0)
				len = snprintf(one, sizeof(one), "%hu ", kip6fwp->fw_number);
			else
				len = snprintf(one, sizeof(one), "%hu.%u ", kip6fwp->fw_number, subnumber);
			if (total_len + len > TXT_LEN) {
				syslog(LOG_INFO, "  ** %s", txt);
				total_len = 0;
			}
			strcpy(txt + total_len, one);
			total_len += len;
		}
		if (total_len != 0)
			syslog(LOG_INFO, "  ** %s", txt);
	}
	return 0;
}

/*
 * Remove IPv6 Firewall support: close descriptor for opened socket.
 */
void
kip6fw_close(void)
{
	syslog(LOG_INFO, "remove IPv6 Firewall support");
	if (sd != -1) {
		if (close(sd) < 0)
			syslog(LOG_ERR, "kip6fw_close: close(FD %d): %m", sd);
		sd = -1;
	}
	free(kip6fw);
}

#endif /* USE_FreeBSD_IP6FW */
